---
package.description = """
x ---y
"""
---

// Test that hyphens are allowed inside frontmatters if there is some
// non-whitespace character preceding them.
//@check-pass

#![feature(frontmatter)]

fn main() {}
