//! Trait definitions.
//!
//! These are each in their own module so we can also define tests for the core types they're impl'd
//! on in the same module.

pub(crate) mod ct_assign;
pub(crate) mod ct_eq;
pub(crate) mod ct_find;
pub(crate) mod ct_gt;
pub(crate) mod ct_lookup;
pub(crate) mod ct_lt;
pub(crate) mod ct_neg;
pub(crate) mod ct_select;
