/*
 * Copyright (c) Edward Thomson.  All rights reserved.
 *
 * This file is part of ntlmclient, distributed under the MIT license.
 * For full terms and copyright information, and for third-party
 * copyright information, see the included LICENSE.txt file.
 */

#ifndef PRIVATE_UNICODE_BUILTIN_H__
#define PRIVATE_UNICODE_BUILTIN_H__

#include <locale.h>
#include <iconv.h>

#include "ntlmclient.h"

struct ntlm_unicode_ctx {
};

#endif /* PRIVATE_UNICODE_BUILTIN_H__ */
